import datetime
import logging

from django.contrib.auth.models import User, AnonymousUser
from testfixtures import LogCapture

from automationcommon.models import set_local_user, Audit, ModelChangeMixin, clear_local_user, LOCAL_USER_WARNING
from automationcommon.tests.utils import UnitTestCase


class FakeModel:
    pk = 1

    def save(self, *args, **kwargs):
        pass

    def delete(self, *args, **kwargs):
        pass


class TestModel(ModelChangeMixin, FakeModel):

    def __init__(self, *args, **kwargs):
        self.fields = {
            'name': 'the round window',
            'description': "it's round",
        }
        super(TestModel, self).__init__(*args, **kwargs)

    @property
    def _dict(self):
        return self.fields.copy()


class ModelsTests(UnitTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create(username="it123")

    def setUp(self):
        set_local_user(self.user)
        self.test_model = TestModel()

    def test_audit_no_user(self):

        clear_local_user()

        with LogCapture(level=logging.INFO) as log_capture:

            # test
            self.test_model.fields.update({'description': "it's a round window"})
            self.test_model.save()

            # check
            self.assertEqual(0, Audit.objects.count())
            log_capture.check((
                'automationcommon', 'WARNING',
                "Don't know who made this change: (model=TestModel:1, field=description, "
                "old='it's round', new='it's a round window')"
            ), (
                'automationcommon', 'WARNING', LOCAL_USER_WARNING
            ))

    def test_audit_single_change(self):

        start = datetime.datetime.now()

        # test
        self.test_model.fields.update({'description': "it's a round window"})
        self.test_model.save()

        # check
        self.assertEqual(1, Audit.objects.count())
        audit = Audit.objects.all().first()
        self.assertLessEqual(start, audit.when)
        self.assertEqual(self.user, audit.who)
        self.assertEqual('TestModel', audit.model)
        self.assertEqual(1, audit.model_pk)
        self.assertEqual('description', audit.field)
        self.assertEqual("it's round", audit.old)
        self.assertEqual("it's a round window", audit.new)

    def test_audit_anonymous_user(self):

        set_local_user(AnonymousUser())

        # test
        self.test_model.fields.update({'description': "it's a round window"})
        self.test_model.save()

        # check
        self.assertEqual(1, Audit.objects.count())
        self.assertIsNone(Audit.objects.all().first().who)

    def test_audit_multiple_change(self):

        # test
        self.test_model.fields = {
            'name': 'the square window',
            'description': "no wait, it's actually square!",
        }
        self.test_model.save()

        # check
        self.assertEqual(2, Audit.objects.count())
        audits = Audit.objects.all().order_by('field')
        self.assertEqual('description', audits[0].field)
        self.assertEqual("it's round", audits[0].old)
        self.assertEqual("no wait, it's actually square!", audits[0].new)
        self.assertEqual('name', audits[1].field)
        self.assertEqual("the round window", audits[1].old)
        self.assertEqual("the square window", audits[1].new)

    def test_audit_delete(self):

        # test
        self.test_model.delete()

        # check
        self.assertEqual(2, Audit.objects.count())
        audits = Audit.objects.all().order_by('field')
        self.assertEqual('description', audits[0].field)
        self.assertEqual("it's round", audits[0].old)
        self.assertIsNone(audits[0].new)
        self.assertEqual('name', audits[1].field)
        self.assertEqual("the round window", audits[1].old)
        self.assertIsNone(audits[1].new)

    def tearDown(self):
        clear_local_user()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Audit',
            fields=[
                ('id', models.AutoField(primary_key=True, verbose_name='ID', serialize=False, auto_created=True)),
                ('when', models.DateTimeField(auto_now=True)),
                ('model', models.CharField(max_length=64)),
                ('model_pk', models.IntegerField()),
                ('field', models.CharField(max_length=64)),
                ('old', models.CharField(blank=True, max_length=255, null=True)),
                ('new', models.CharField(blank=True, max_length=255, null=True)),
                ('who', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, null=True)),
            ],
        ),
    ]



import logging
import threading

from django.contrib.auth.models import User
from django.db import models
from django.forms import model_to_dict

LOGGER = logging.getLogger('automationcommon')

LOCAL_USER_WARNING = """
    Use automationcommon.models.set_local_user() to set the user to be used in the audit trail or 
    automationcommon.middleware.RequestUserMiddleware if you are in the context of a webapp.
"""


class Creatable(models.Model):
    """
    An abstract Model encapsulating an entity that can be created
    """

    # creator of the entity
    creator = models.ForeignKey(User, related_name="+")
    # when the entity was created
    creation_date = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True


class Audit(models.Model):
    """
    A model that defines an audit record for a change to any django model

    Attributes:
        when      when the change was made
        who       who made the change (if null, then the user was anomymous)
        model     the changed model name
        model_pk  the changed model primary key
        field     the changed model's field
        old       the changed field's original value
        new       the changed field's updated value
    """
    when = models.DateTimeField(auto_now=True)

    who = models.ForeignKey(User, null=True, blank=True)

    model = models.CharField(max_length=64)

    model_pk = models.IntegerField()

    field = models.CharField(max_length=64)

    old = models.CharField(max_length=255, null=True, blank=True)

    new = models.CharField(max_length=255, null=True, blank=True)


# A thread local object used for binding the user currenting updating the model to the thread.
_thread_local = threading.local()


def set_local_user(user):
    """
    Bind's a user to the current thread to be used for the audit trail

    :param user: user model
    """
    _thread_local.user = user


def get_local_user():
    """
    :return: The user for the local thread's request
    """
    return _thread_local.user if hasattr(_thread_local, 'user') else None


def clear_local_user():
    """
    Clear's the user from the current thread
    """
    _thread_local.user = None


class ModelChangeMixin(object):
    """
    A model mixin that tracks changes to model fields' values and saves an Audit record per changed field
    when the model is saved. Based ModelDiffMixin on here:
    https://stackoverflow.com/questions/1355150/django-when-saving-how-can-you-check-if-a-field-has-changed
    """
    def __init__(self, *args, **kwargs):
        super(ModelChangeMixin, self).__init__(*args, **kwargs)
        self.__initial = self._dict

    @property
    def _dict(self):
        """
        :return: a dict of the model's fields and their current values
        """
        # NOTE: an internal attribute has been used when introspecting the model.
        return model_to_dict(self, fields=[field.name for field in self._meta.fields])

    @property
    def diffs(self):
        """
        :return: An array of any changed fields. Each item is a sequence: (field_name, (original_value, updated_value))
        """
        d1 = self.__initial
        d2 = self._dict
        return [(k, (v, d2[k])) for k, v in d1.items() if v != d2[k]]

    def save(self, *args, **kwargs):
        """
        Saves model, created an Audit record per changed field, and resets the initial state.
        """
        creating = self.pk is None
        super(ModelChangeMixin, self).save(*args, **kwargs)
        # Don't audit new records
        if not creating:
            request_user = get_local_user()
            for diff in self.diffs:
                if request_user:
                    Audit.objects.create(
                        who=None if request_user.is_anonymous() else request_user,
                        model=self.__class__.__name__,
                        model_pk=self.pk,
                        field=diff[0],
                        old=diff[1][0], new=diff[1][1]
                    )
                else:
                    LOGGER.warning("Don't know who made this change: (model=%s:%s, field=%s, old='%s', new='%s')" % (
                        self.__class__.__name__, self.pk, diff[0], diff[1][0], diff[1][1]
                    ))
                    LOGGER.warning(LOCAL_USER_WARNING)

        self.__initial = self._dict

    def delete(self, *args, **kwargs):
        """
        Created an Audit record per field with 'new' set to None and deletes the model.
        """
        request_user = get_local_user()
        if request_user:
            for field, value in self.__initial.items():
                if field != 'id' and value:
                    Audit.objects.create(
                        who=None if request_user.is_anonymous() else request_user,
                        model=self.__class__.__name__,
                        model_pk=self.pk,
                        field=field, old=value,
                    )
        else:
            LOGGER.warning("Don't know deleted this: (model=%s:%s)" % (self.__class__.__name__, self.pk))
            LOGGER.warning(LOCAL_USER_WARNING)
        super(ModelChangeMixin, self).delete(*args, **kwargs)



from automationcommon.models import clear_local_user, set_local_user


class RequestUserMiddleware(object):
    """
    Middleware that simply set's the request.user to be used for the audit trail.
    """
    @classmethod
    def process_request(cls, request):
        set_local_user(request.user)

    @classmethod
    def process_response(cls, request, response):
        """
        Clear the user to minimise the chance of a user being wrongly assigned.
        """
        clear_local_user()
        return response

    @classmethod
    def process_exception(cls, request, exception):
        """
        Clear the user to minimise the chance of a user being wrongly assigned.
        """
        clear_local_user()



from django.contrib import admin
from django.contrib.admin import ModelAdmin

from automationcommon.models import Audit


class AuditAdmin(ModelAdmin):
    list_display = ('when', 'who', 'model', 'model_pk', 'field', 'old', 'new')


admin.site.register(Audit, AuditAdmin)
